/***************************************************************************
    smb4kmounter.cpp  -  The core class, that mounts shares.
                             -------------------
    begin                : Die Jun 10 2003
    copyright            : (C) 2003 by Alexander Reinholdt
    email                : dustpuppy@mail.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// Qt includes
#include <qapplication.h>
#include <qdir.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qregexp.h>

// KDE includes
#include <kapplication.h>

// system includes
#include <unistd.h>
#include <sys/types.h>

// Application specific includes
#include "smb4kmounter.h"
#include "../smb4k.h"


Smb4KMounter::Smb4KMounter( QObject *parent, const char *name ) : QObject( parent, name )
{
  m_config = kapp->config();
  
  m_proc = new KProcess();
  m_proc->setUseShell( true );

  m_state = Idle;
  m_lastJob = false;

  // The timer for the queue.
  m_timer1 = new QTimer( this );
  m_timer1->start( 50, false );

  // The timer for the checks.
  m_timer2 = new QTimer( this );
  m_timer2->start( 1000, false );

  m_working = false;

  // Check, whether anything is already mounted.
  m_queue.append( QString( "%1:" ).arg( Import ) );

  connect( m_proc, SIGNAL( processExited( KProcess * ) ), SLOT( slotProcessExited( KProcess * ) ) );
  connect( m_proc, SIGNAL( receivedStdout( KProcess *, char *, int ) ), SLOT( slotReceivedStdout( KProcess *, char *, int ) ) );
  connect( m_proc, SIGNAL( receivedStderr( KProcess *, char *, int ) ), SLOT( slotReceivedStderr( KProcess *, char *, int ) ) );
  connect( m_timer1, SIGNAL( timeout() ), SLOT( init() ) );
  connect( m_timer2, SIGNAL( timeout() ), SLOT( check() ) );
}


Smb4KMounter::~Smb4KMounter()
{
  delete m_proc;
}


/***************************************************************************
   Aborts any action of the mounter.
***************************************************************************/


void Smb4KMounter::abort()
{
  if ( m_proc->isRunning() )
    m_proc->kill();

  m_queue.clear();
}


/***************************************************************************
   Imports all shares, that are mounted externally.
***************************************************************************/


void Smb4KMounter::import()
{ 
  // /proc/mounts is read to import the externally mounted shares.
  QFile file( "mounts" );
  QDir *dir = new QDir();

  if ( QFile::exists( "/proc/mounts" ) )
  {
    dir->setPath( "/proc" );
    QDir::setCurrent( dir->path() );
  }
  else
    qWarning( "ERROR [Smb4KMounter::import()]: Could not find /proc/mounts! That's bad.\nPlease compile your kernel with support for the /proc filesystem and try again!" );

  if ( !dir->path().isEmpty() )
  {
    if ( file.open( IO_ReadOnly ) )
    {
      QStringList data;

      QTextStream ts( &file );
      while( !ts.atEnd() )
      {
        QString line = ts.readLine().stripWhiteSpace();
        if ( line.contains( "smbfs", false ) != 0 )
        {
          QString mounted = line.section( "smbfs", 0, 0 ).stripWhiteSpace();
          mounted = mounted.replace( " ", ":" );
          // Get the uid and gid of the user, who mounted this share and append them to the list.
          QString uid = line.section( "uid=", 1, 1 ).section( ",", 0, 0 ).stripWhiteSpace();
          QString gid = line.section( "gid=", 1, 1 ).section( ",", 0, 0 ).stripWhiteSpace();

          // Now put everything in one string.
          mounted.append( QString( ":%1:%2" ).arg( uid ).arg( gid ) );
          
          // Append the share to the list that will be passed to the share view.          
          data.append( mounted );
        }
      }
      file.close();

      // Copy the data:
      m_mountedShares = data;
    }
    else
      qWarning( QString( "ERROR [Smb4KMounter::import()]: Could not open %1/%2!" ).arg( dir->path() ).arg( file.name() ) );
  }

  emit imported( m_mountedShares );

  // The mounter finished. Tell it, that's not busy anymore and start the timer
  // again.
  m_working = false;
  m_timer1->start( 50, false );
}


/***************************************************************************
   This function is run on exit of the whole program. It cares about the
   last second actions, that are to be done.
***************************************************************************/


void Smb4KMounter::exit()
{
  // First tell the program, that this will be the last job
  // of the mounter for this session.
  m_lastJob = true;  

  // Clean up!
  if ( m_cleanUp )
  {
    // Now set the path, where the progam has to clean up.
    QDir *dir = new QDir();
    dir->setPath( m_defaultPath );

    // Get the contents of the default directory and remove everything from
    // the list we do not need.
    QStringList dirs = dir->entryList( QDir::Dirs, QDir::DefaultSort );
    dirs.remove( "." );
    dirs.remove( ".." );

    // Now, remove all directories, that are present in the base directory
    // except those that are to remain mounted.
    for ( QStringList::ConstIterator it = dirs.begin(); it != dirs.end(); ++it )
    {
      dir->cd( *it );
      QStringList subdirs = dir->entryList( QDir::Dirs, QDir::DefaultSort );
      subdirs.remove( "." );
      subdirs.remove( ".." );

      // Make sure, that no directory of a mounted share will be touched. The
      // admins of the file servers will get after you, if something will
      // accidentally be deleted!
      for ( QStringList::ConstIterator i = subdirs.begin(); i != subdirs.end(); ++i )
        if ( m_mountedShares.grep( dir->absPath().append( "/" ).append( *i ) ).isEmpty() )
          dir->rmdir( *i );

      dir->cdUp();
      dir->rmdir( *it );
    }
  }
  // Unmount shares.
  if ( m_unmountAll )
    unmountAll();
}


/***************************************************************************
   Mounts a share. (Public part)
***************************************************************************/


void Smb4KMounter::mountShare( const QString &workgroup, const QString &host, const QString &share )
{
  QString item( QString( "%1:%2:%3:%4" ).arg( Mount ).arg( workgroup ).arg( host ).arg( share ) );
  m_queue.append( item );
}



/***************************************************************************
   Mounts a share. (Private part)
***************************************************************************/


void Smb4KMounter::mount( const QString &workgroup, const QString &host, const QString &share )
{
  m_workgroup = workgroup;
  m_host = host;
  
  QString mountpoint;
  if ( share.stripWhiteSpace().contains( " " ) )
  {
    m_share = QString( "'%1'" ).arg( share );
    mountpoint = share;
    mountpoint = mountpoint.replace( QRegExp( " " ), "_" );
  }
  else
    m_share = share;
  m_path = m_defaultPath;

  // Test, whether the share is already mounted. We test for the mount point
  // and the share. This has to be done, because the user might change the 
  // default directory.
  QString test_mp( QString( "%1%2/%3" ).arg( m_path ).arg( m_host ).arg( m_share ) );
  QString test_share( QString( "//%1/%2" ).arg( m_host ).arg( m_share ) );
  bool alreadyMounted = false;
    
  QStringList first_try = m_mountedShares.grep( test_mp, false );
    
  if ( !first_try.isEmpty() )
  {
    uint index( 0 );
    while ( index < first_try.count() )
    {
      if ( first_try[index].section( ":", -3, -3 ).stripWhiteSpace().length() == test_mp.length() )
      {
        alreadyMounted = true;
        break;
      }
      index++;
    }
  }
  else
  {
    QStringList second_try = m_mountedShares.grep( test_share, false );
    if ( !second_try.isEmpty() )
    {
      uint index( 0 );
      while ( index < second_try.count() )
      {
        if ( second_try[index].section( ":", -2, -2 ).toInt() == (int)getuid() )
        {
          alreadyMounted = true;
          break;
        }
        index++;
      }
    }
  }
    
  // Start the mount process.
  if ( alreadyMounted == true )
  {
    emit error( QString::null, AlreadyMounted );
    m_timer1->start( 50, false );
  }
  else
  {
    QDir *dir = new QDir();
    dir->setPath( m_path );
    if ( !dir->exists() )
      dir->mkdir( m_path, true );
       
    // If the user wants the generated subdirectories to be lower case,
    // then we have to care about this right here:
    if( !m_lowercase )
    {
      m_path.append( host );
      dir->setPath( m_path );
      if ( !dir->exists() )
        dir->mkdir( m_path, true );
      if ( !mountpoint.isEmpty() )
        m_path.append( "/" ).append( mountpoint );
      else
        m_path.append( "/" ).append( share );
      dir->setPath( m_path );
      if ( !dir->exists() )
        dir->mkdir( m_path, true );
    }
    else
    {
      m_path.append( host.lower() );
      dir->setPath( m_path );
      if ( !dir->exists() )
        dir->mkdir( m_path, true );
      if ( !mountpoint.isEmpty() )
        m_path.append( "/" ).append( mountpoint.lower() );
      else
        m_path.append( "/" ).append( share.lower() );
      dir->setPath( m_path );
      if ( !dir->exists() )
        dir->mkdir( m_path, true );
    }
    
    // Get the authentication data:
    QStringList auth = sapp->passwdReader()->read( workgroup, host, share );

    QString command( QString( "smbmount //%1/%2 %3 -o workgroup=%4" ).arg( KProcess::quote( m_host ) ).arg( KProcess::quote( share ) ).arg( KProcess::quote( m_path ) ).arg( KProcess::quote( workgroup ) ) );

    // If a password and username is defined, append it to the command.
    if ( !auth[0].isEmpty() )
		{
      command.append( QString( ",username=\"%1\"" ).arg( auth[0] ) );
      command.append( QString( ",password=\"%1\"" ).arg( auth[1] ) );
		}
    else
      command.append( ",guest" );

    m_proc->operator<<( command );
    startProcess( Mount );
  }
}


/****************************************************************************
   Unmounts a share. (Public part)
****************************************************************************/


void Smb4KMounter::unmountShare( const QString &mountpoint )
{
  QString item( QString( "%1:%2" ).arg( Unmount ).arg( mountpoint ) );
  m_queue.append( item );
}


/***************************************************************************
   Unmounts a share. (Private part)
***************************************************************************/

void Smb4KMounter::unmount( const QString &mountpoint )
{
  if ( !mountpoint.isEmpty() )
  {
    m_path = mountpoint;
    QString command( "smbumount " );
    command.append( mountpoint );

    m_proc->operator<<( command );
    startProcess( Unmount );
  }
}


/***************************************************************************
   Unmounts all shares at once. (Public part)
***************************************************************************/


void Smb4KMounter::unmountAllShares()
{
  QString item( QString( "%1" ).arg( UnmountAll ) );
  m_queue.append( item );
}


/***************************************************************************
   Unmounts all shares at once.
***************************************************************************/


void Smb4KMounter::unmountAll()
{
  QString command;
  for ( QStringList::ConstIterator it = m_mountedShares.begin(); it != m_mountedShares.end(); ++it )
  {
    QString path = (*it).section( ":", -3, -3 ).stripWhiteSpace();
    command.append( "smbumount " ).append( KProcess::quote( path ) ).append( " ; " );
  }
  command.truncate( command.length() - 2 );

  // If the user wants to clean up the standard directory on exit, we
  // want to do him this favour.
  // ATTENTION: DO NOT change '&&' into ';'! Just don't.
  if ( m_lastJob && m_cleanUp && !m_mountedShares.isEmpty() )
    command.append( " && rm -rf " ).append( KProcess::quote( m_defaultPath ) ).append( "*" );

  m_proc->operator<<( command );
  
  startProcess( UnmountAll );
}


/***************************************************************************
   Starts any process.
***************************************************************************/


void Smb4KMounter::startProcess( int state )
{
  emit running( true, state );
  
  m_buffer = QString::null;
  m_state = state;
  if ( m_state != Import )
    QApplication::setOverrideCursor( waitCursor );
  m_proc->start( KProcess::NotifyOnExit, KProcess::AllOutput );
}


/***************************************************************************
   Ends any process. This functions tells the mounter what to do 
   afterwards.
***************************************************************************/


void Smb4KMounter::endProcess()
{
  switch ( m_state )
  {
    case Mount:
      processMount();
      break;
    case Unmount:
      processUnmount();
      break;
    case UnmountAll:
      processUnmountAll();
      break;
    default:
      break;
  }
  m_state = Idle;
  m_path = QString::null;
  QApplication::restoreOverrideCursor();
  m_proc->clearArguments();
  emit running( false, m_state );

  m_timer1->start( 50, false );
}


/***************************************************************************
   Process mounts.
***************************************************************************/


void Smb4KMounter::processMount()
{
  if ( m_buffer.contains( "failed", true ) == 0 && m_buffer.contains( "ERR", true ) == 0 )
  {
    // If the share does contain spaces, then we have to replace them by
    // underscores and we also have to remove the single quotation marks.
    if ( m_share.contains( "'" ) == 2 )
    {
      m_share = m_share.replace( QRegExp( "'" ), "" );
      if ( m_share.contains( " " ) != 0 )
        m_share = m_share.replace( QRegExp( " " ), "_" );
    }

    QString share( QString( "//%1/%2:%3:%4:%5" ).arg( m_host ).arg( m_share ).arg( m_path ).arg( (int)getuid() ).arg( (int)getgid() ) );

    // Keep also this share in mind.
    m_mountedShares.append( share );

    emit mounted( share );
  }
  else
  {
    if ( m_buffer.contains( "ERRbadpw" ) != 0 || m_buffer.contains( "ERRnoaccess" ) != 0 )
    {
      // If the user supplied auth information, we will retry mounting.
      if ( sapp->passwdReader()->askpass( m_workgroup, m_host, m_share ) )
      {
        QString item( QString( "%1:%2:%3:%4" ).arg( Mount ).arg( m_workgroup ).arg( m_host ).arg( m_share ) );
        m_queue.append( item );
      }
    }
    else
      emit error( m_buffer, MountError );
  }

  m_workgroup = QString::null;
  m_host = QString::null;
  m_share = QString::null;
}


/***************************************************************************
   Process unmounts.
***************************************************************************/


void Smb4KMounter::processUnmount()
{
  // FIXME: This is a potential problem. What if the unmount
  // process returns some debugging info like the mount process
  // under SuSE? Need more info.
  if ( m_buffer.isEmpty() )
  {
    // Since the mount point is unique, the mount point is greped
    // and used to determine the share that has to be removed from 
    // m_mountedShares.
    
    QString entry = m_mountedShares.grep( m_path, true ).first();
    m_mountedShares.remove( entry );
    
    emit unmounted( entry );
  }
  else
    emit error( m_buffer, UnmountError );
}


/***************************************************************************
   What to do after all shares were unmounted at once.
***************************************************************************/


void Smb4KMounter::processUnmountAll()
{
  // If everything went fine, the buffer should have
  // length == 0. Then, just clear the list of mounted
  // shares and emit the signal.
  if ( m_buffer.isEmpty() )
  {
    m_mountedShares.clear();
    QStringList data;
    emit unmountedAll( m_mountedShares );
  }
  // If something went wrong, remove all shares that
  // could be unmounted from the list and emit the error
  // signal.
  else
  {
    uint index( 0 );
    while ( index < m_mountedShares.count() )
    {
      QString path = m_mountedShares[index].section( ":", -3, -3 );
      
      // It turned out, that using QStringList::remove() here wont't work.
      // Thus, all unmounted shares are set to NULL here and will be
      // removed later.      
      if ( m_buffer.contains( path ) == 0 )
        m_mountedShares[index] = QString::null;
      index++;
    }
    // Remove all NULL entries.
    m_mountedShares.remove( QString::null );
    
    emit unmountedAll( m_mountedShares );
    emit error( m_buffer, UnmountAllError );
  }
}


/***************************************************************************
   Read the options.
***************************************************************************/


void Smb4KMounter::readOptions()
{
  m_config->setGroup( "Mount Options" );
  m_defaultPath = m_config->readEntry( "Default Path", QDir::homeDirPath().append( "/smb4k/" ) );
  // Set this variable to true, so that unmountAll() knows, that it has
  // to put a 'rm -rf *' after the unmounting command when exiting.
  m_cleanUp = m_config->readBoolEntry( "Clean Up", true );
  m_unmountAll = m_config->readBoolEntry( "Unmount All", false );
  m_lowercase = m_config->readBoolEntry( "Force Lowercase", false );
}


/////////////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATIONS
/////////////////////////////////////////////////////////////////////////////

/****************************************************************************
   Internal slots.
****************************************************************************/


void Smb4KMounter::slotProcessExited( KProcess * )
{
  endProcess();
}


void Smb4KMounter::slotReceivedStdout( KProcess *, char *buf, int len )
{
  m_buffer.append( QString::fromLocal8Bit( buf, len ) );
}


void Smb4KMounter::slotReceivedStderr( KProcess *, char *buf, int len )
{
  m_buffer.append( QString::fromLocal8Bit( buf, len ) );
}


void Smb4KMounter::init()
{
  if ( !m_working && !m_queue.isEmpty() )
  {
    // Tell the mounter, that it is busy.
    m_working = true;
    // Stop the timer.
    m_timer1->stop();

    // Process the first entry.
    int todo = m_queue[0].section( ":", 0, 0 ).toInt();

    // Now process the list.
    if ( todo == Import )
      import();
    else if ( todo == Mount )
    {
      QString workgroup = m_queue[0].section( ":", 1, 1 );
      QString host = m_queue[0].section( ":", 2, 2 );
      QString share = m_queue[0].section( ":", 3, 3 );
      
      mount( workgroup, host, share );
    }
    else if ( todo == Unmount )
    {
      QString mountpoint = m_queue[0].section( ":", 1, 1 );
      unmount( mountpoint );
    }
    else if ( todo == UnmountAll )
      unmountAll();

    // Remove the first entry of the list and set the mounter
    // in non-working mode.
    m_queue.remove( m_queue.begin() );
    m_working = false;
  }
}


void Smb4KMounter::check()
{
  // Just append what has to be checked to the queue.
  m_queue.append( QString( "%1:" ).arg( Import ) );
}

#include "smb4kmounter.moc"
