/***************************************************************************
    smb4kmounter.h  -  The core class, that mounts shares.
                             -------------------
    begin                : Die Jun 10 2003
    copyright            : (C) 2003 by Alexander Reinholdt
    email                : dustpuppy@mail.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SMB4KMOUNTER_H
#define SMB4KMOUNTER_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// Qt includes
#include <qobject.h>
#include <qstringlist.h>
#include <qtimer.h>

// KDE includes
#include <kprocess.h>
#include <kconfig.h>


/** This is a core class. It mounts the shares. */

class Smb4KMounter : public QObject
{
  Q_OBJECT

  public:
    /** The constructor. */
    Smb4KMounter( QObject *parent = 0, const char *name = 0 );
    /** The destructor.
    */
    ~Smb4KMounter();
    /** Aborts a running process.
    */
    void abort();
    /** Enumeration for the state of the process. The members are also
        used in the results() signal as return values. */
    enum State{ Import, Mount, Unmount, UnmountAll, Idle };
    /** Enumeration, that determines the kind of error, that occurred. */
    enum Error{ AlreadyMounted, MountError, UnmountError, UnmountAllError };
    /** Does everything, that has to be done on exit. Retruns TRUE
        if everything has been accomplished. */
    void exit();
    /** Reads the options. */
    void readOptions();
    /** Unmounts a share. */
    void unmountShare( const QString &mountpoint );
    /** Unmounts all shares. */
    void unmountAllShares();
    /** Mounts a share. */
    void mountShare( const QString &workgroup, const QString &host, const QString &share );

  signals:
    /** Is emitted, if an error occured. */
    void error( const QString &output, int errorValue );
    /** Emits the state, the mounter is in. */
    void running( bool on, int state );
    /** Emits the imported data. */
    void imported( const QStringList &data );
    /** Is emitted, if a share has been unmounted. The share widget knows the share
        already; it is highlighted. */
    void unmounted( const QString &share );    
    /** Emits the mounted share data. */
    void mounted( const QString &data );
    /** Is emitted, if all shares should have been unmounted. The string list is
        empty, if the unmount process ended successfully. */
    void unmountedAll( const QStringList &shares );

  protected:
    /** Starts a process. It takes an interger, that determines,
        which process has to be started.
    */
    void startProcess( int state );
    /** Is called, when the process ended.
    */
    void endProcess();
    /** Finishes the mounting of shares.
    */
    void processMount();
    /** Finishes the unmounting of a single share.
    */
    void processUnmount();
    /** Finishes the unmounting of all shares.
    */
    void processUnmountAll();
    
  protected slots:
    /** Is called, when the process exits.
    */
    void slotProcessExited( KProcess * );
    /** Is called, if output is received at Stdout.
    */
    void slotReceivedStdout( KProcess *, char *buf, int len );
    /** Is called, if output is received a Stderr.
    */
    void slotReceivedStderr( KProcess *, char *buf, int len );
    /** Initiates the processing of the queue. */
    void init();
    /** Writes the checks into the queue. */
    void check();

  private:
    /** Imports mounted shares. */
    void import();
    /** Mounts a selected share. */
    void mount( const QString &workgroup, const QString &host, const QString &share );
    /** Unmounts the selected item. */
    void unmount( const QString &mountpoint );
    /** Unmounts all shares a once. */
    void unmountAll();
    /** Holds the list of mounted shares. */
    QStringList m_mountedShares;
    /** The KProcess object. */
    KProcess *m_proc;
    /** The KConfig objects. With its help the options are read. */
    KConfig *m_config;
    /** The buffer. */
    QString m_buffer;
    /** Holds the mount point of a share. */
    QString m_path;
    QString m_workgroup;
    QString m_host;
    QString m_share;
    /** The default path where Smb4K mounts the shares. */
    QString m_defaultPath;
    int m_state;
    bool m_lastJob;
    bool m_cleanUp;
    bool m_unmountAll;
    /** The queue, where the incoming requests are stored. */
    QStringList m_queue;
    /** The timer for the queue. */
    QTimer *m_timer1;
    /** The timer for all the checks that should be done time by time. */
    QTimer *m_timer2;
    /** Determines, whether the mounter is running or not. */
    bool m_working;
    /** Holds whether the generate subdirectories have to be lower case
        or not. */
    bool m_lowercase;
};
#endif
