/***************************************************************************
    smb4kscanner.cpp  -  The network scan core class of Smb4K.
                             -------------------
    begin                : Sam Mai 31 2003
    copyright            : (C) 2003 by Alexander Reinholdt
    email                : dustpuppy@mail.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// Qt includes
#include <qapplication.h>

// KDE includes
#include <klocale.h>
#include <kapplication.h>

// Application specific includes.
#include "smb4kscanner.h"
#include "../smb4k.h"


Smb4KScanner::Smb4KScanner( QObject *parent, const char *name ) : QObject( parent, name )
{
  m_proc = new KProcess();
  m_proc->setUseShell( true );

  m_state = Idle;
  m_workgroup = QString::null;
  m_host = QString::null;
  m_share = QString::null;
  m_path = QString::null;

  m_config = kapp->config();

  m_timer = new QTimer( this );
  m_timer->start( 50, false );

  m_working = false;

  m_queue.append( QString( "%1:" ).arg( Init ) );

  connect( m_proc, SIGNAL( receivedStdout( KProcess *, char *, int ) ), SLOT( slotReceivedStdout( KProcess *, char *, int ) ) );
	connect( m_proc, SIGNAL( processExited( KProcess* ) ), SLOT( slotProcessExited( KProcess * ) ) );
  connect( m_proc, SIGNAL( receivedStderr( KProcess *, char *, int ) ), SLOT( slotReceivedStderr( KProcess *, char *, int ) ) );
	connect( m_timer, SIGNAL( timeout() ), SLOT( start() ) );
	connect( this, SIGNAL( running( bool ) ), SLOT( slotRunning( bool ) ) );}

Smb4KScanner::~Smb4KScanner()
{
  delete m_proc;
}


/****************************************************************************
   These functions initialize/abort any action, that's done by the scanner.
****************************************************************************/


void Smb4KScanner::init()
{
  QString command;

  // Use nmblookup or query host to obtain the browse list?
  if ( m_method == "nmblookup" )
  {
    command.append( "nmblookup -M -- - | grep '<01>' | awk '{print $1}'" );

    // Query the WINS server, if available.
    if ( m_wins.isEmpty() )
      command.append( " | xargs nmblookup -A | sed -e '/<00>/!d'" );
    else
      command.append( QString( " | xargs nmblookup -R -U %1 -A | sed -e '/<00>/!d'" ).arg( m_wins ) );

    m_proc->operator<<( command );
    startProcess( Groups );
  }
  else if ( m_method == "host" )
  {
    command.append( "smbclient -d1 -N -L " ).append( KProcess::quote( m_hostToStartWith ) );

		m_proc->operator<<( command );
    startProcess( QueryHost );
  }
}


void Smb4KScanner::abort()
{
  if ( m_proc->isRunning() )
    m_proc->kill();

  // FIXME: We can make this configurable, whether the complete list
  // of actions should be cleared or only the current process should be
  // killed.
  m_queue.clear();
  emit killed();
}


/****************************************************************************
   Gets the preview of a share. (public part)
****************************************************************************/


void Smb4KScanner::getPreview( const QString &workgroup, const QString &host, const QString &share, const QString &path )
{
  QString item( QString( "%1:%2:%3:%4:%5" ).arg( Preview ).arg( workgroup ).arg( host ).arg( share ).arg( path ) );
	m_queue.append( item );
}


/****************************************************************************
   Gets the preview of a share. (private part)
****************************************************************************/


void Smb4KScanner::preview( const QString &workgroup, const QString &host, const QString &share, const QString &path )
{
  m_workgroup = workgroup;
  m_host = host;
  m_share = share;
  m_path = path;

  QString command;
  QStringList auth = sapp->passwdReader()->read( workgroup, host, share );

  // Assemble command:
  command = QString( "smbclient //%1/%2 -d1 -W %3 -N -c 'ls" ).arg( KProcess::quote( host ) ).arg( KProcess::quote( share ) ).arg( KProcess::quote( workgroup ) );
  if ( !path.isEmpty() )
	{
    // To be able to display directories, that contain umlauts, we have to do
    // two things:
    // (1) replace all slashes by backslashes;
    // (2) convert the directory name to local 8 bit.
    QString proc_path = path;
    proc_path = proc_path.replace( QChar( '/' ), QChar( '\\' ) ).local8Bit();
    command.append( " \"" ).append( proc_path ).append( "*\"" );
  }
  command.append( "'" );

  if ( !auth[0].isEmpty() )
	{
    command.append( QString( " -U \"%1\"" ).arg( auth[0] ) );
		if ( !auth[1].isEmpty() )
		  command.append( QString( "%\"%2\"" ).arg( auth[1] ) );
  }

  m_proc->operator<<( command );
  startProcess( Preview );
}


/****************************************************************************
   Scans for workgroup members. (public part)
****************************************************************************/


void Smb4KScanner::getGroupMembers( const QString &workgroup, const QString &master )
{
  QString item( QString( "%1:%2:%3" ).arg( Hosts ).arg( workgroup ).arg( master ) );
	m_queue.append( item );
}


/****************************************************************************
   Scans for workgroup members. (private part)
****************************************************************************/


void Smb4KScanner::scanWGMembers( const QString &workgroup, const QString &master )
{
  m_workgroup = workgroup;

  QString command;

  // Scan for the host list.
  command.append( "smbclient -d1 -N -L " ).append( KProcess::quote( master ) );
  command.append( " -W " ).append( KProcess::quote( workgroup ) );

	m_proc->operator<<( command );
  startProcess( Hosts );
}


/****************************************************************************
   Scans for shares on a selected host. (public part)
****************************************************************************/


void Smb4KScanner::getShares( const QString &workgroup, const QString &host )
{
  QString item( QString( "%1:%2:%3" ).arg( Shares ).arg( workgroup ).arg( host ) );
	m_queue.append( item );
}


/****************************************************************************
   Scans for shares on a selected host. (private part)
****************************************************************************/


void Smb4KScanner::scanShares( const QString &workgroup, const QString &host )
{
  m_workgroup = workgroup;
  m_host = host;

  QString command;
  QStringList auth = sapp->passwdReader()->read( workgroup, host, QString::null );

	// Assemble the scan command.
  command.append( "smbclient -d1 -N -L " ).append( KProcess::quote( host ) );
  command.append( " -W " ).append( KProcess::quote( workgroup ) );

  if ( !auth[0].isEmpty() )
  {
		command.append( QString( " -U \"%1\"" ).arg( auth[0] ) );
		if ( !auth[1].isEmpty() )
		  command.append( QString( "%\"%2\"" ).arg( auth[1] ) );
	}

  m_proc->operator<<( command );
  startProcess( Shares );
}


/****************************************************************************
   Gets more info on a selected host. (public part)
****************************************************************************/


void Smb4KScanner::getInfo( const QString &workgroup, const QString &host )
{
  QString item( QString( "%1:%2:%3" ).arg( Info ).arg( workgroup ).arg( host ) );
	m_queue.append( item );
}


/****************************************************************************
   Gets more info on a selected host. (private part)
****************************************************************************/


void Smb4KScanner::scanForInfo( const QString &workgroup, const QString &host )
{
  QString command;

  // Assemble the scan command.
  command.append( "smbclient -d1 -N -L " ).append( KProcess::quote( host ) );
  command.append( " -W " ).append( KProcess::quote( workgroup ) );

  m_proc->operator<<( command );
  startProcess( Info );
}


/****************************************************************************
   Searches for a host. (public part)
****************************************************************************/


void Smb4KScanner::makeSearch( const QString &host )
{
  QString item( QString( "%1:%2" ).arg( Search ).arg( host ) );
  m_queue.append( item );
}


/****************************************************************************
   Searches for a host. (private part)
****************************************************************************/


void Smb4KScanner::search( const QString &host )
{
  m_host = host;

  QString command;

  // If there is a WINS server, it will be queried for the host.
  if ( !m_wins.isEmpty() )
    command = QString( "nmblookup -R -U %1 %2 -S | grep '<00>'| sed -e 's/<00>.*//'" ).arg( m_wins ).arg( m_host );

		// If we have none, the program does a normal look-up.
  else
    command = QString( "nmblookup %1 -S | grep '<00>'| sed -e 's/<00>.*//'" ).arg( m_host );

  m_proc->operator<<( command );
  startProcess( Search );
}


/****************************************************************************
   Starts the scanner.
****************************************************************************/


void Smb4KScanner::startProcess( int state )
{
  m_state = state;
  m_buffer = QString::null;
  QApplication::setOverrideCursor( waitCursor );
  m_proc->start( KProcess::NotifyOnExit, KProcess::AllOutput );
}


/****************************************************************************
   Tell the program what it has to do.
****************************************************************************/


void Smb4KScanner::endProcess()
{
  switch ( m_state )
  {
    case Groups:
      processGroups();
      break;
    case QueryHost:
      processQueryHost();
      break;
    case Hosts:
      processHosts();
      break;
    case Shares:
      processShares();
      break;
    case Info:
      processInfo();
      break;
    case Preview:
      processPreview();
      break;
    case Search:
      processSearch();
      break;
    default:
      break;
  }
  m_state = Idle;
  m_workgroup = QString::null;
  m_host = QString::null;
  m_share = QString::null;
  m_path = QString::null;
  QApplication::restoreOverrideCursor();
  m_proc->clearArguments();
  emit running( false );

  // Start the timer again.
  m_timer->start( 50, false );
}


/****************************************************************************
   Process workgroups.
****************************************************************************/

void Smb4KScanner::processGroups()
{
  QStringList list = QStringList::split( '\n', m_buffer, false );
  QStringList workgroupList;
  
  int hosts = 0;

  // Put the masters and the workgroups in a QStringList:
  for ( QStringList::ConstIterator it = list.begin(); it != list.end(); ++it )
  {
    // Master
    if ( (*it).contains( "<00>" ) && !(*it).contains( "<GROUP>" ) )
    {
      if ( hosts < 1 )
      {
        workgroupList.append( (*it).left( (*it).find( "<00>" ) ).stripWhiteSpace() );
				hosts++;
      }
      else
        continue;
    }
    // Group
    else if ( (*it).contains( "<00>" ) && (*it).contains( "<GROUP>" ) && hosts == 1 )
		{
      hosts = 0;
      workgroupList.append( (*it).left( (*it).find( "<00>" ) ).stripWhiteSpace());
		}
  }

  // Emit the list.   
  emit workgroups( workgroupList );
}


void Smb4KScanner::processQueryHost()
{
  QStringList list = QStringList::split( '\n', m_buffer, false );
  QStringList workgroupList;

  // Get the line starting with Workgroup.
  uint index( 0 );
  while( index < list.count() )
  {
    if ( list[index++].stripWhiteSpace().startsWith( "Workgroup" ) )
      break;
    else
      continue;
  }

  index++;

  while ( index < list.count() )
  {
    QString line = list[index].stripWhiteSpace();
    QString master, workgroup;
    master = line.section( "   ", -1, -1 ).stripWhiteSpace();
    workgroup = line.section( "   ", 0, -2 ).stripWhiteSpace();
    if ( workgroup.isEmpty() )
    {
      workgroupList.append( i18n( "Unknown" ) );
      workgroupList.append( line.stripWhiteSpace() );
    }
    else
    {
      workgroupList.append( master );
      workgroupList.append( workgroup );
    }
    index++;
  }

  // Emit the list of workgroups and masters.
  emit workgroups( workgroupList );
}


/****************************************************************************
   Process the memeber list of a workgroup.
****************************************************************************/


void Smb4KScanner::processHosts()
{
  QStringList list = QStringList::split( '\n', m_buffer, false );

  if ( list.grep( "Connection to" ).count() != 0 )
  {
    QString errmsg = list.grep( "Connection to" ).first().stripWhiteSpace();
    emit scan_error( errmsg );
  }
  else
  {
    // Get the begin of the host listing.
    uint index( 0 );
    for ( ; index < list.count(); index++ )
      if ( list[index].stripWhiteSpace().startsWith( "Server" ) )
        break;

    index += 2;

    // Process the host list.
    QString line;
    QStringList data;
    while ( index < list.count() )
    {
      line = list[index++].stripWhiteSpace();
      if ( line.startsWith( "Workgroup" ) )
        break;
      else
      {
        // Append the line to the return data.
        data.append( line );      
      }
    }
    emit members( m_workgroup, data );
  }
}


/****************************************************************************
   Process the share list of a host.
****************************************************************************/


void Smb4KScanner::processShares()
{
  QStringList list = QStringList::split( '\n', m_buffer, false );

  // First process the info:
  uint index( 0 );

  // The program hasn't emitted any info yet.
  bool emittedInfo = false;

  for ( ; index < list.count(); index++ )
  {
    // We want the OS and the Server info as well.
    if ( list[index].stripWhiteSpace().startsWith( "Domain" ) || list[index].stripWhiteSpace().startsWith( "OS" ) )
		{
      // The OS string.
      QString os = list[index].right( list[index].length() - list[index].find( "OS", 0 ) );
			if ( !os.isEmpty() )
        os = os.section( '[', 1, 1 ).section( ']', 0, 0 );
      // The Server string.
      QString server = list[index].right( list[index].length() - list[index].find( "Server", 0 ) );
			if ( !server.isEmpty() )
        server = server.section( '[', 1, 1 ).section( ']', 0, 0 );
      // Emit the results and set the emittedInfo to true.
      emittedInfo = true;
      emit info( server, os );
    }
    else if ( list[index].stripWhiteSpace().startsWith( "Sharename" ) )
      break;
  }
  // If there was no OS and server info, we should at least emit
  // two null strings.
  if ( !emittedInfo )
    emit info( QString::null, QString::null );

  // Now do the rest:
  if ( list.grep( "NT_STATUS" ).count() != 0 || list.grep( "Connection to" ).count() != 0 || list.grep( "tree connect failed:" ).count() != 0 )
	{
    if ( list.grep( "NT_STATUS" ).count() != 0  )
    {
      QString errmsg = list.grep( "NT_STATUS" ).first().stripWhiteSpace().section( " ", -1, -1 );

			// The error output of smbclient is a little bit inconsistent:
			if ( errmsg.contains( "NT_STATUS" ) == 0 )
			  errmsg = list.grep( "NT_STATUS" ).first().stripWhiteSpace().section( " ", 0, 0 );
			if ( errmsg == "NT_STATUS_ACCESS_DENIED" || errmsg == "NT_STATUS_LOGON_FAILURE" )
			{
        if ( sapp->passwdReader()->askpass( m_workgroup, m_host, QString::null )
)       {
          QString item( QString( "%1:%2:%3" ).arg( Shares ).arg( m_workgroup ).arg( m_host ) );
					m_queue.append( item );
        }
      }
      else
        emit scan_error( errmsg );
    }
    else if ( list.grep( "tree connect failed:" ).count() != 0 )
    {
      if ( sapp->passwdReader()->askpass( m_workgroup, m_host, QString::null ) )
      {
        QString item( QString( "%1:%2:%3" ).arg( Shares ).arg( m_workgroup ).arg( m_host ) );
				m_queue.append( item );
      }
    }
    else
    {
      QString errmsg = list.grep( "Connection to" ).first().stripWhiteSpace();
      emit scan_error( errmsg );
    }
  }
  else if ( list.grep( "Error returning browse list:" ).count() != 0 && list.grep( "NT_STATUS" ).count() == 0 )
	{
    QString errmsg = list.grep( "Error returning browse list:" ).first().section( ":", 1, 1 ).stripWhiteSpace();
		emit scan_error( errmsg );
  }
  else
  {
    index += 2;

    // Process the share listing.
    QString line;
    QStringList data;
    while ( index < list.count() )
    {
      line = list[index++].stripWhiteSpace().simplifyWhiteSpace();
      if ( line.startsWith( "Server" ) )
        break;
      else if ( line.startsWith( "Anonymous" ) )
        continue;
      else
      {
        // Put the share with its type and comment into the list.
        data.append( line );
      }
    }
    emit shares( m_workgroup, m_host, data );
  }
}


/****************************************************************************
   Process the information about a host.
****************************************************************************/


void Smb4KScanner::processInfo()
{
  QStringList data = QStringList::split( '\n', m_buffer, false );

  // Search for the right line.
  uint index( 0 );
  while ( index < data.count() )
  {
    if ( data[index].stripWhiteSpace().startsWith( "Domain" ) || data[index].stripWhiteSpace().startsWith( "OS" ) )
		{
      // The OS string.
      QString os = data[index].right( data[index].length() - data[index].find( "OS", 0 ) );
			if ( !os.isEmpty() )
        os = os.section( '[', 1, 1 ).section( ']', 0, 0 );
      // The Server string.
      QString server = data[index].right( data[index].length() - data[index].find( "Server", 0 ) );
			if ( !server.isEmpty() )
        server = server.section( '[', 1, 1 ).section( ']', 0, 0 );
      // Emit the results.
      emit info( server, os );
      break;
    }
    else if ( data[index].contains( "Connection to", true ) != 0 )
    {
      emit info( QString::null, QString::null );
      emit scan_error( data[index].stripWhiteSpace() );
    }
    else
      emit info( QString::null, QString::null );
    index++;
  }
}


/****************************************************************************
   Process the preview.
****************************************************************************/


void Smb4KScanner::processPreview()
{
  QStringList list = QStringList::split( '\n', m_buffer, false );

  // Now do the rest:
  if ( list.grep( "NT_STATUS" ).count() != 0 || list.grep( "Connection to" ).count() != 0 )
	{
    if ( list.grep( "NT_STATUS" ).count() != 0  )
    {
      QString errmsg = list.grep( "NT_STATUS" ).first().stripWhiteSpace().section( " ", 0, 0 );
			// The error output of smbclient is a little bit inconsistent:
			if ( errmsg.contains( "NT_STATUS" ) == 0 )
			  errmsg = list.grep( "NT_STATUS" ).first().stripWhiteSpace().section( " ", -1, -1 );
			if ( errmsg == "NT_STATUS_ACCESS_DENIED" || errmsg == "NT_STATUS_LOGON_FAILURE" )
			{
        if ( sapp->passwdReader()->askpass( m_workgroup, m_host, m_share ) )
        {
          QString item( QString( "%1:%2:%3:%4:%5" ).arg( Preview ).arg( m_workgroup ).arg( m_host ).arg( m_share ).arg( m_path ) );
					m_queue.append( item );
				}
      }
      else if ( errmsg == "NT_STATUS_NO_SUCH_FILE" )
        emit prev_error( errmsg );
    }
    else
    {
      QString errmsg = list.grep( "Connection to" ).first().stripWhiteSpace();
      emit prev_error( errmsg );
    }
  }
  else if ( list.grep( "Error returning browse list:" ).count() != 0 && list.grep( "NT_STATUS" ).count() == 0 )
	{
    QString errmsg = list.grep( "Error returning browse list:" ).first().section( ":", 1, 1 ).stripWhiteSpace();
		emit scan_error( errmsg );
  }
  else
    emit previewResult( list );
}


/****************************************************************************
   Process the search data.
****************************************************************************/


void Smb4KScanner::processSearch()
{
  QStringList data = QStringList::split( "\n", m_buffer.stripWhiteSpace(), false );
  // Blank the first line. The contained information is useless and
  // disturbing (for now).
  data[0] = QString::null;

  // If the host wasn't found, the entries in the list are empty.
  if ( data.count() > 1 && !data[1].isEmpty() )
    emit searchResult( data.last().stripWhiteSpace(), data.grep( m_host, false ).first().stripWhiteSpace() );
	else
    emit searchResult( QString::null, QString::null );
}


/****************************************************************************
   Read the options.
****************************************************************************/


void Smb4KScanner::readOptions()
{
  // Determine, which look-up method should be used:
  m_config->setGroup( "Browse Options" );
  m_method = m_config->readEntry( "Browse List", "nmblookup" );
  bool wins = m_config->readBoolEntry( "WINS Support", false );

  // Get the WINS server.
  if ( wins )
    m_wins = m_config->readEntry( "WINS Server", QString::null );

  // Get the host to query for the browse list.
  if ( m_method == "host" || wins == true )
    m_hostToStartWith = m_config->readEntry( "Query Host", m_wins );
  else
    m_hostToStartWith = QString::null;
}


/////////////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATIONS
/////////////////////////////////////////////////////////////////////////////


/****************************************************************************
   Internal slots.
****************************************************************************/

void Smb4KScanner::slotReceivedStdout( KProcess *, char *buf, int len )
{
  m_buffer.append( QString::fromLocal8Bit( buf, len ) );
}


void Smb4KScanner::slotProcessExited( KProcess * )
{
  endProcess();
}


void Smb4KScanner::slotReceivedStderr( KProcess *, char  *buf, int len )
{
  m_buffer.append( QString::fromLocal8Bit( buf, len ) );
}


void Smb4KScanner::start()
{
  if ( !m_working && m_queue.count() != 0 )
  {
    // Tell the program, that the scanner is running again.
    emit running( true );

    // Stop the timer.
    m_timer->stop();

    // Start processing...
    int todo = m_queue[0].section( ":", 0, 0 ).toInt();
    QString workgroup, host;

    if ( todo == Init )
    {
      init();
    }
    else if ( todo == Hosts )
    {
      workgroup = m_queue[0].section( ":", 1, 1 );
      host = m_queue[0].section( ":", 2, 2 );
      scanWGMembers( workgroup, host );
    }
    else if ( todo == Shares )
    {
      workgroup = m_queue[0].section( ":", 1, 1 );
      host = m_queue[0].section( ":", 2, 2 );
      scanShares( workgroup, host );
    }
    else if ( todo == Info )
    {
      workgroup = m_queue[0].section( ":", 1, 1 );
      host = m_queue[0].section( ":", 2, 2 );
      scanForInfo( workgroup, host );
    }
    else if ( todo == Preview )
    {
      workgroup = m_queue[0].section( ":", 1, 1 );
      host = m_queue[0].section( ":", 2, 2 );
      QString share = m_queue[0].section( ":", 3, 3 );
      QString path = m_queue[0].section( ":", 4, 4 );
      preview( workgroup, host, share, path );
    }
    else if ( todo == Search )
    {
      host = m_queue[0].section( ":", 1, 1 );
      search( host );
    }

    // Remove the first entry from the list.
    m_queue.remove( m_queue.begin() );

    // ATTENTION: The browser window sends todo == Shares and todo == GetInfo
    // requests back-on-back. The latter are redundant, because the necessary
    // info is also provided by processShares(). In the following we will remove
    // these entries.
    if ( m_queue[0].contains( QString( "%1" ).arg( Info ) ) && m_queue[0].contains( workgroup ) && m_queue[0].contains( host ) )
		{
      m_queue.remove( m_queue.begin() );
		}
  }
}


/****************************************************************************
   External slots.
****************************************************************************/


void Smb4KScanner::slotRunning( bool on )
{
  m_working = on;
}


void Smb4KScanner::slotInit()
{
  QString item( QString( "%1:" ).arg( Init ) );
  m_queue.append( item );
}


#include "smb4kscanner.moc"

