/***************************************************************************
    smb4kscanner.h  -  The network scan core class of Smb4K.
                             -------------------
    begin                : Sam Mai 31 2003
    copyright            : (C) 2003 by Alexander Reinholdt
    email                : dustpuppy@mail.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef SMB4KSCANNER_H
#define SMB4KSCANNER_H

#ifndef HAVE_CONFIG_H
#include <config.h>
#endif

// Qt includes
#include <qobject.h>
#include <qstringlist.h>
#include <qtimer.h>

// KDE includes
#include <kprocess.h>
#include <kconfig.h>

/** This is the core class, that communicates with the network. All look-up
    stuff is done here. */

class Smb4KScanner : public QObject
{
  Q_OBJECT

  public:
    /** The constructor. */
    Smb4KScanner( QObject *parent = 0, const char *name = 0 );
    /** The destructor. */
    ~Smb4KScanner();

    /** Aborts the network scan. */
    void abort();
    /** Reads the options. */
    void readOptions();
    /** Scan for the shares on a selected host. */
    void getShares( const QString &workgroup, const QString &host );
    /** Scans for workgroup members. */
    void getGroupMembers( const QString &, const QString & );
    /** Get more info about a share (i.e. server and OS string, etc.). */
    void getInfo( const QString &workgroup, const QString &host );
    /** Searches for a givn host. */
    void makeSearch( const QString &host );   
    /** Get the preview of a share. */
    void getPreview( const QString &workgroup, const QString &host, const QString &share, const QString &path );
    
  signals:
    /** This signal is emitted, when the scanner starts and ends. */
    void running( bool );
    /** Emits the list of workgroups and masters, that were found. */
    void workgroups( const QStringList &list );
    /** Emits the list of workgroup member hosts. */
    void members( const QString &workgroup, const QStringList &list );
    /** Emits the list of shares. */
    void shares( const QString &workgroup, const QString &host, const QStringList &list );
    /** This signal provides info about a certain host. It passes the server
        and the OS string. */
    void info( const QString &serverString, const QString &osString );
    /** This signal is emitted, if the process was killed. */
    void killed();
    /** Is emitted by the preview process. */
    void previewResult( const QStringList &data );
    /** Is emitted, when the results of a network search are to be passed. */
    void searchResult( const QString &workgroup, const QString &host );
    /** Emits the error message, that occurred during scan. */
    void scan_error( const QString &message );
    /** Emits the errors, that occurred during preview. */
    void prev_error( const QString &message );

  protected:
    /** Starts the scan process. */
    void startProcess( int state );
    /** Is called, when the process ends. */
    void endProcess();
    /** Processes the output of the network group scan. */
    void processGroups();
    /** Processes the output of the master host scan. */
    void processHosts();
    /** Processes the output of the host scan. */
    void processShares();
    /** Processes the output of the scan for the OS and Server
        string. */
    void processInfo();
    /** Processes the output you get after querying the WINS
        server. */
    void processQueryHost();
    /** Processes the output of the preview request. */
    void processPreview();
    /** Precesses the output of a search request. */
    void processSearch();


  protected slots:
    /** Is called, if something is received via stdout. */
    void slotReceivedStdout( KProcess *, char *buf, int len );
    /** Is called, when the KProcess exited. */
    void slotProcessExited( KProcess * );
    /** Is called, if something is received via stderr. */
    void slotReceivedStderr( KProcess *, char  *buf, int len );
    /** Starts processing the queue. */
    void start();
    /** This slot is called by the running() signal. */
    void slotRunning( bool on );
    /** This slot is called, if the network search should be reinitiated. */
    void slotInit();
    
    
  private:
    /** Initializes the network scan. */
    void init();
    /** Scans the group master for the group members. */
    void scanWGMembers( const QString &workgroup, const QString &master );
    /** Scans the chosen host for its shares.
    */
    void scanShares( const QString &workgroup, const QString &host );
    /** Scans a given host for its OS and Server string. */
    void scanForInfo( const QString &workgroup, const QString &host );
    /** Searches for a given host. */
    void search( const QString &host );
    /** Gets a preview of a share. */
    void preview( const QString &workgroup, const QString &host, const QString &share, const QString &path );
    /** The KConfig object. */
    KConfig *m_config;
    /** Internal enumeration. */
    enum TODO{ Groups, QueryHost, Hosts, Shares, Info, Preview, Search, Init, Idle };
    /** The KProcess object. */
    KProcess *m_proc;
    /** Determines the state the process is in. */
    int m_state;
    /** The buffer. */
    QString m_buffer;
    /** Holds the look-up method string. */
    QString m_method;
    /** Holds the host name / host IP, that is needed by init. */
    QString m_hostToStartWith;
    /** Holds the current workgroup, that was passed by Smb4KBrowserWidget. */
    QString m_workgroup;
    /** Holds the current host, that was passed by Smb4KBrowserWidget. */
    QString m_host;
    /** Holds the current share. */
    QString m_share;
    /** Holds the current path (for preview). */
    QString m_path;
    /** "Buffers" the incoming requests and stores them until they are processed. */
    QStringList m_queue;
    /** The internal timer. */
    QTimer *m_timer;
    /** The IP address or host name of the WINS server. */
    QString m_wins;
    /** Is set to true, if the scanner is running. */
    bool m_working;
};

#endif
