/***************************************************************************
    smb4kstarter.cpp  -  Stuff that's carried out on start-up.
                             -------------------
    begin                : Mit Apr 30 2003
    copyright            : (C) 2003 by Alexander Reinholdt
    email                : dustpuppy@mail.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// Qt includes
#include <qdir.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qlayout.h>
#include <qtimer.h>
#include <qregexp.h>

// KDE includes.
#include <kcmdlineargs.h>
#include <klocale.h>
#include <kiconloader.h>

// system includes
#include <unistd.h>
#include <sys/types.h>
#include <sys/utsname.h>

// application specific includes
#include "smb4kstarter.h"
#include "../smb4k.h"
#include "smb4kpasswdreader.h"

Smb4KStarter::Smb4KStarter( QObject *parent, const char *name ) : QObject( parent, name )
{
  m_config = kapp->config();
  
  // Get the users UID and GID.
  m_uid = getuid();
  m_gid = getgid();

  m_proc = new KProcess();
  m_proc->setUseShell( true );

  splashScreen();
  
  connect( m_proc, SIGNAL( processExited( KProcess * ) ), SLOT( slotProcessExited( KProcess * ) ) );}


Smb4KStarter::~Smb4KStarter()
{
  delete m_proc;
}


/****************************************************************************
   Brings up the splash screen.
****************************************************************************/


void Smb4KStarter::splashScreen()
{
  m_splashScreen = new QFrame( 0, 0, Qt::WStyle_Splash|Qt::WDestructiveClose );
  m_splashScreen->setFrameShape( QFrame::PopupPanel );
  m_splashScreen->setFrameShadow( QFrame::Raised );  m_splashScreen->setLineWidth( 2 );
  QGridLayout *layout = new QGridLayout( m_splashScreen );
  layout->setMargin( 10 );
  layout->setSpacing( 10 );

  QLabel *picLabel = new QLabel( m_splashScreen );
  picLabel->setPixmap( DesktopIcon( "smb4k" ) );
  picLabel->adjustSize();

  // Text label:
  m_infoLabel = new KSqueezedTextLabel( i18n( "Starting Smb4K..." ), m_splashScreen );
	m_infoLabel->setFrameShape( QFrame::PopupPanel );
  m_infoLabel->setFrameShadow( QFrame::Sunken );
  m_infoLabel->setFixedHeight( picLabel->height() );

  // Progress bar:
  m_progress = new KProgress( 100, m_splashScreen, 0, 0 );
  m_progress->setMargin( 0 );
  m_progress->setFixedHeight( 10 );
  m_progress->setTextEnabled( false );
  m_progress->setProgress( 0 );

  layout->addWidget( picLabel, 0, 0, 0 );
  layout->addWidget( m_infoLabel, 0, 1, 0 );
  layout->addMultiCellWidget( m_progress, 1, 1, 0, 1 );

  m_splashScreen->setFixedHeight( 2 * picLabel->height() );
  m_splashScreen->setFixedWidth( 10 * picLabel->width() );

  m_splashScreen->setGeometry( QApplication::desktop()->width()/2 - m_splashScreen->width()/2,
                              QApplication::desktop()->height()/2 - m_splashScreen->height()/2,
                              m_splashScreen->width()/2,
															m_splashScreen->height()/2 );
  m_splashScreen->show();
}


/****************************************************************************
   Takes care of everything that has to be done before the main window
   comes up and shows it afterwards.
****************************************************************************/


void Smb4KStarter::init()
{
  // Get the OS name:
  m_infoLabel->setText( i18n( "Detecting operating system..." ) );
  detectOS();
  m_progress->setProgress( 20 );
  
  // Does the host use a WINS server? Let's find out.
  m_infoLabel->setText( i18n( "Detecting WINS server..." ) );
  checkForWINS();
  m_progress->setProgress( 40 );

  // Detect the already mounted shares. This list is necessary
  // for the mounting of recently used shares later on.
  m_infoLabel->setText( i18n( "Checking for mounted shares..." ) );
  m_mounted = checkForShares();
  m_progress->setProgress( 60 );
  
  // Get the shares, that are to be mounted.
  m_infoLabel->setText( i18n( "Looking for shares to mount..." ) );
  m_list = getSharesToMount();
  m_progress->setProgress( 80 );
  
  // Mount the shares, if there are any.  
  if ( !m_list.isEmpty() )
  {
    m_infoLabel->setText( i18n( "Mounting shares. Please wait..." ) );
    mount();
    // In this case the rest is done in Smb4KStarter::slotProcessExited()
  }
  else
  {
    m_progress->setProgress( 100 );
    m_infoLabel->setText( i18n( "Welcome to Smb4K!" ) );
    runProgram();
  }
}


/****************************************************************************
   Gets the shares that are to be mounted.
****************************************************************************/


const QStringList Smb4KStarter::getSharesToMount()
{
  QStringList list;
  
  if ( m_config->hasGroup( "Mount Options" ) )
  {
    m_config->setGroup( "Mount Options" );
    if( m_config->readBoolEntry( "Mount Recent", false ) )
    {
      if ( m_config->hasGroup( "Recently Mounted Shares" ) )
      {
        m_config->setGroup( "Recently Mounted Shares" );
        QString temp;
        for ( int index = 0; ; index++ )
        {
          temp = m_config->readEntry( QString( "%1" ).arg( index ), 
QString::null );
          if ( !temp.isEmpty() )
          {
            // We have to append the user's UID and GID:
            temp.append( QString( ":%1:%2" ).arg( getuid() ).arg( getgid() ) );
            
            // OK, this seems unlogical, but we do not want to mount
            // a share more than once:
            if ( !list.isEmpty() )
              list.remove( temp );
            list.append( temp );
          }
          else
            break;
        }        
      }      
    }
  }
  
  return list;
}


/****************************************************************************
   Mounts the shares.
****************************************************************************/


void Smb4KStarter::mount()
{
  // The command variable.
  QString command;
  
  // Get the default path.
  m_config->setGroup( "Mount Options" );
  m_path = m_config->readEntry( "Default Path", QDir::homeDirPath().append( "/smb4k/" ) );
	for ( QStringList::ConstIterator it = m_list.begin(); it != m_list.end(); ++it )
	{
    if ( m_mounted.grep( (*it).stripWhiteSpace() ).count() == 0 )
    {
      // Make the directories, where the shares will be mounted.
      QString path = m_path;
      QDir *dir = new QDir();
      dir->setPath( path );
      if ( !dir->exists() )
        dir->mkdir( path, true );

      // If the generated subdirectories should be lowercase, we have to take
      // care of this right now:
      if ( !m_config->readBoolEntry( "Force Lowercase", false ) )
      {
        path.append( (*it).section( '/', 2, 2 ) );
        dir->setPath( path );
        if ( !dir->exists() )
          dir->mkdir( path, true );
        path.append( "/" ).append( (*it).section( '/', 3, 3 ).section( ":", 0, 0 ) );
				dir->setPath( path );
        if ( !dir->exists() )
          dir->mkdir( path, true );
      }
      else
      {
        path.append( (*it).section( '/', 2, 2 ).lower() );
        dir->setPath( path );
        if ( !dir->exists() )
          dir->mkdir( path, true );
        path.append( "/" ).append( (*it).section( '/', 3, 3 ).section( ":", 0, 0 ).lower() );
				dir->setPath( path );
        if ( !dir->exists() )
          dir->mkdir( path, true );
      }

      // Get the the login and the password.
      Smb4KPasswdReader *reader = new Smb4KPasswdReader();
      QStringList data = reader->read( (*it).section( "/", 2, 2 ), (*it).section( "/", 3, 3 ) );

      // Assemble the command. Special attention has to be given to shares, that were stored
      // containing underscores.
      QString share = (*it).section( ":", 0, 0 );

      // (1) Take the share name as it is.
      command.append( QString( "smbmount %1 %2" ).arg( KProcess::quote( share ) ).arg( KProcess::quote( path ) ) );
      if ( !data[0].isEmpty() )
        command.append( QString( " -o username=\"%1\",password=\"%2\"" ).arg( data[0] ).arg( data[1] ) );
		  else
        command.append( " -o guest" );
      command.append( " ; " );

      // (2) Replace the underscores by spaces.
      if ( share.section( '/', 3, 3 ).contains( "_" ) != 0 )
      {
        QString mod_share = QString( "%1/%2" ).arg( share.section( '/', 0, 2 ) ).arg( share.section( '/', 3, 3 ).replace( QRegExp( "_" ), " " ) );
        QStringList mod_data = reader->read( mod_share.section( "/", 2, 2 ), mod_share.section( "/", 3, 3 ) );

        command.append( QString( "smbmount %1 %2" ).arg( KProcess::quote( mod_share ) ).arg( KProcess::quote( path ) ) );
        if ( !mod_data[0].isEmpty() )
        command.append( QString( " -o username=\"%1\",password=\"%2\"" ).arg( mod_data[0] ).arg( mod_data[1] ) );
		    else
          command.append( " -o guest" );
        command.append( " ; " );
      }
    }
  }

  if ( !command.isEmpty() )
  {
    command.truncate( command.stripWhiteSpace().length() - 2 );
    m_proc->operator<<( command );
    QApplication::setOverrideCursor( waitCursor );
    m_proc->start( KProcess::NotifyOnExit, KProcess::NoCommunication );
  }
  else
    runProgram();
}


/****************************************************************************
   Runs the program.
****************************************************************************/


void Smb4KStarter::runProgram()
{
  m_progress->setProgress( 100 );

  // Kill the busy cursor.
  m_proc->clearArguments();
  QApplication::restoreOverrideCursor();

  m_infoLabel->setText( i18n( "Welcome to Smb4K!" ) );

  if ( kapp->isRestored() )
  {
    RESTORE( Smb4KApp );
  }
  else
  {
    Smb4KApp *smb4k = new Smb4KApp();
    smb4k->show();

    KCmdLineArgs *args = KCmdLineArgs::parsedArgs();

    args->clear();
  }
  QTimer::singleShot( 1500, this, SLOT( slotClose() ) );
}


/****************************************************************************
   Detect the operating system.
****************************************************************************/


void Smb4KStarter::detectOS()
{
  m_config->setGroup( "System" );

  // Initialize system info stuff.
  struct utsname system_info;
  uname( &system_info );

  m_config->writeEntry( "Name", QString( "%1" ).arg( system_info.sysname ) );
  m_config->writeEntry( "Version", QString( "%1" ).arg( system_info.release ) );
}


/****************************************************************************
   Determines whether there is a WINS server configured.
****************************************************************************/


void Smb4KStarter::checkForWINS()
{
  QString wins;

  QFile file( "smb.conf" );
  QDir *dir = new QDir();

  if ( QFile::exists( "/etc/smb.conf" ) )
  {
    dir->setPath( "/etc" );
    QDir::setCurrent( dir->path() );
  }
  else if ( QFile::exists( "/etc/samba/smb.conf" ) )
  {
    dir->setPath( "/etc/samba" );
    QDir::setCurrent( dir->path() );
  }
  else if ( QFile::exists( "/usr/local/etc/smb.conf" ) )
  {
    dir->setPath( "/usr/local/etc" );
    QDir::setCurrent( dir->path() );
  }
  else
    qWarning( "ERROR [Smb4KStarter::checkForWINS()]: Haven't found smb.conf! Continuing..." );
  if ( !dir->path().isEmpty() )
  {
    if ( file.open( IO_ReadOnly ) )
    {
      QTextStream ts( &file );
      while( !ts.atEnd() )
      {
        QString line = ts.readLine().stripWhiteSpace().upper();
        if ( line.startsWith( "WINS SERVER" ) )
        {
          wins = line.section( "=", 1, 1 ).stripWhiteSpace();
          break;
        }
        else if ( line.startsWith( "WINS SUPPORT" ) )
        {
          if ( line.section( "=", 1, 1 ).stripWhiteSpace() == "YES" || line.section( "=", 1, 1 ).stripWhiteSpace() == "TRUE" )
					  wins = "127.0.0.1";
				  break;
        }
      }
      file.close();
    }
    else
      qWarning( QString( "ERROR [Smb4KStarter::checkForWINS()]: Could not open %1/%2!" ).arg( dir->path() ).arg( file.name() ) );  }

  // If there is a WINS server defined, put it into the config file
  // and enable the WINS support.
  if ( !wins.isEmpty() )
  {
    m_config->setGroup( "Browse Options" );
    m_config->writeEntry( "WINS Support", true );
    m_config->writeEntry( "WINS Server", wins );


    // If a WINS server is defined, Smb4K will query that one by
    // default to get the browse list:
    if ( !m_config->hasKey( "Browse List" ) )
      m_config->writeEntry( "Browse List", "host" );
    if ( !m_config->hasKey( "Query Host" ) )
      m_config->writeEntry( "Query Host", wins );
    m_config->sync();
  }
  else
  {
    m_config->setGroup( "Browse Options" );
    m_config->writeEntry( "WINS Support", false );
    if ( m_config->hasKey( "WINS Server" ) );
      m_config->deleteEntry( "WINS Server" );
    m_config->sync();
  }
}


/****************************************************************************
   Checks for already mounted shares.
****************************************************************************/


const QStringList Smb4KStarter::checkForShares()
{
  QStringList shares;

  QDir *dir = new QDir();
  QFile file( "mounts" );
  dir->setPath( QString::null );

  if ( QFile::exists( "/proc/mounts" ) )
  {
    dir->setPath( "/proc" );
    QDir::setCurrent( dir->path() );
  }
  else
    qWarning( QString( "ERROR [Smb4KStarter::checkForShares()]: Could not find %1/%2! Continuing..." ).arg( dir->path() ).arg( file.name() ) );
  if ( !dir->path().isEmpty() )
  {
    if ( file.open( IO_ReadOnly ) )
    {
      QTextStream ts( &file );
      while( !ts.atEnd() )
      {
        QString line = ts.readLine().stripWhiteSpace();
        if ( line.contains( "smbfs", false ) != 0 )
        {
          // Share name.
          QString share = line.section( " ", 0, 0 ).stripWhiteSpace();
          // Share's uid.
          QString uid = line.section( "uid=", 1, 1 ).section( ",", 0, 0 ).stripWhiteSpace();
					// Share's gid.
          QString gid = line.section( "gid=", 1, 1 ).section( ",", 0, 0 ).stripWhiteSpace();
          shares.append( QString( "%1:%2:%3" ).arg( share ).arg( uid ).arg( gid ) );
				}
      }
      file.close();
    }
    else
      qWarning( QString( "ERROR [Smb4KStarter::checkForShares()]: Could not open %1/%2!" ).arg( dir->path() ).arg( file.name() ) );  }

  return shares;
}


/////////////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATIONS
/////////////////////////////////////////////////////////////////////////////


void Smb4KStarter::slotProcessExited( KProcess * )
{
  runProgram();
}


void Smb4KStarter::slotClose()
{
  m_splashScreen->close();
}


#include "smb4kstarter.moc"

