/***************************************************************************
    smb4ksharewidget.cpp  -  The widget, that displays the mounted shares.
                             -------------------
    begin                : Sam Mr 1 2003
    copyright            : (C) 2003 by Alexander Reinholdt
    email                : dustpuppy@mail.berlios.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// Qt includes
#include <qdir.h>
#include <qregexp.h>

// KDE includes
#include <klocale.h>
#include <kprocess.h>
#include <kpopupmenu.h>
#include <kmessagebox.h>
#include <kapplication.h>

// system includes
#include <unistd.h>
#include <sys/types.h>

// application specific includes
#include "smb4ksharewidget.h"
#include "../core/smb4kmounter.h"
#include "../smb4k.h"


Smb4KShareWidget::Smb4KShareWidget( QWidget *parent, const char *name ) : KIconView( parent, name )
{
  m_config = kapp->config();
  
  setFrameShape( KIconView::PopupPanel );
  setFrameShadow( KIconView::Sunken );
  setSelectionMode( KIconView::Single );
  setResizeMode( KIconView::Adjust );
  setShowToolTips( true );
  setAutoArrange( true );
  setSorting( true, true );
  
  // Get the UID and GID of the user:
  m_uid = (int)getuid();
  m_gid = (int)getgid();
  
  m_read = false;

  // Internal connections
  connect( this, SIGNAL( executed( QIconViewItem * ) ), SLOT( slotItemExecuted( QIconViewItem * ) ) );
  connect( this, SIGNAL( mouseButtonPressed( int, QIconViewItem *, const QPoint & ) ), SLOT( slotMouseButtonPressed(  int, QIconViewItem *, const QPoint & ) ) );
  connect( this, SIGNAL( selectionChanged( QIconViewItem * ) ), SLOT( slotSelectionChanged( QIconViewItem * ) ) );

  // External connections
  connect( sapp->mounter(), SIGNAL( imported( const QStringList & ) ), this, SLOT( slotImported( const QStringList & ) ) );
  connect( sapp->mounter(), SIGNAL( mounted( const QString & ) ), this, SLOT( slotMounted( const QString & ) ) );
  connect( sapp->mounter(), SIGNAL( unmounted( const QString & ) ), this, SLOT( slotUnmounted( const QString & ) ) );
  connect( sapp->mounter(), SIGNAL( unmountedAll( const QStringList & ) ), this, SLOT( slotUnmountedAll( const QStringList & ) ) );
  connect( sapp->mounter(), SIGNAL( error( const QString &, int ) ), this, SLOT( slotReceivedError( const QString &, int ) ) );  
}


Smb4KShareWidget::~Smb4KShareWidget()
{
}


void Smb4KShareWidget::exit()
{
  // Save the list of currently mounted shares.
  m_config->setGroup( "Mount Options" );
  if ( m_config->readBoolEntry( "Mount Recent", false ) )
  {
    m_config->deleteGroup( "Recently Mounted Shares", true, false );    
    m_config->setGroup( "Recently Mounted Shares" );
    
    int index = 0;
    
    for ( QIconViewItem *item = this->firstItem(); item; item = item->nextItem() )
    {
      // We want to save the shares that Smb4K's user has mounted, but not
      // those that were mounted by other users.
      if ( m_mountpoint )
      {
        // We have to test for the mount point, because only that one is unique.
        QString entry = m_shares.grep( item->text(), true ).first();
        if ( entry.section( ":", 2, 2 ).stripWhiteSpace().toInt() == m_uid )
          m_config->writeEntry( QString( "%1" ).arg( index, 0, 10 ).stripWhiteSpace(), item->key() );
        else
          continue;
      }
      else
      {
        // We have to test for the mount point, because only that one is unique.
        QString entry = m_shares.grep( item->key(), true ).first();
        if ( entry.section( ":", 2, 2 ).stripWhiteSpace().toInt() == m_uid )
          m_config->writeEntry( QString( "%1" ).arg( index, 0, 10 ).stripWhiteSpace(), item->text() );
        else
           continue;
      }
      index++;
    }
  }
  // Should we clear the icon view? Well, why not...
  clear();

  // Clear the list of mounted shares:
  m_shares.clear();
}


/****************************************************************************
   Reads the options for the share view widget.
****************************************************************************/


void Smb4KShareWidget::readOptions()
{
  // Lock the import function.
  m_read = true;
  
  // Which appearance does the user want?
  m_config->setGroup( "Appearance" );
  m_display = m_config->readEntry( "Display Shares", "icons" );
  if ( m_display == "icons" )
  {
    setItemTextPos( KIconView::Bottom );
    setMaxItemWidth( 150 );
    setArrangement( KIconView::LeftToRight );
    setWordWrapIconText( true );
  }
  else if ( m_display == "list" )
  {
    setItemTextPos( KIconView::Right );
    setMaxItemWidth( 250 );
    setArrangement( KIconView::TopToBottom );
    setWordWrapIconText( false );
  }
  
  m_mountpoint = m_config->readBoolEntry( "Show Mount Point", false );

  // Should all share be shown, that were mounted on the system?
  m_showExternal = m_config->readBoolEntry( "Show All Shares", false );
  
  // Now the items are going to be processed. First clear the whole icon view.
  clear();
  
  // Read the string list and put the items back in the icon view.
  for ( QStringList::ConstIterator it = m_shares.begin(); it != m_shares.end(); ++it )    
  {
    QString share = (*it).section( ":", 0, 0 ).stripWhiteSpace();
    QString path = (*it).section( ":", 1, 1 ).stripWhiteSpace();
    QString share_uid = (*it).section( ":", 2, 2 ).stripWhiteSpace();
    QString share_gid = (*it).section( ":", 3, 3 ).stripWhiteSpace();
    
    // Test whether the share belongs to the user.    
    bool externalMount = false;
    if ( share_uid.toInt() != m_uid && share_gid.toInt() != m_gid )
      externalMount = true;

    // Put the items into the icon view.   
    if ( m_showExternal )
    {      
      if ( !m_mountpoint )
      {
        if ( m_display == "icons" )
        {
          Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, share, "hdd_mount", m_display, externalMount );
          item->setKey( path );
        }
        else if ( m_display == "list" )
        {
          Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, share, "hdd_mount", m_display, externalMount );
          item->setKey( path );          
        }
      }
      else
      {
        if ( m_display == "icons" )
        {
          Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, path, "hdd_mount", m_display, externalMount );
          item->setKey( share );           
        }
        else if ( m_display == "list" )
        {
          Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, path, "hdd_mount", m_display, externalMount );           
          item->setKey( share );
        }
      }
    }
    else
    {
      if ( !m_mountpoint && !externalMount )
      {
        if ( m_display == "icons" )
        {
          Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, share, "hdd_mount", m_display, externalMount );
          item->setKey( path );
        }
        else if ( m_display == "list" )
        {
          Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, share, "hdd_mount", m_display, externalMount );
          item->setKey( path );
        }
      }
      else if ( m_mountpoint && !externalMount )
      {
        if ( m_display == "icons" )
        {
          Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, path, "hdd_mount", m_display, externalMount );
          item->setKey( share );        
        }
        else if ( m_display == "list" )
        {
          Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, path, "hdd_mount", m_display, externalMount );           
          item->setKey( share );
        }
      }    
    }
  }
  
  // Unlock the import function.
  m_read = false;
}


/////////////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATIONS
/////////////////////////////////////////////////////////////////////////////


/****************************************************************************
   Internal slots.
****************************************************************************/


void Smb4KShareWidget::slotItemExecuted( QIconViewItem *item )
{
  if ( item != 0 )
  {
    if ( m_mountpoint )
      kapp->invokeBrowser( item->text() );
    else
      kapp->invokeBrowser( item->key() );
  }
}


void Smb4KShareWidget::slotMouseButtonPressed( int button, QIconViewItem *item, const QPoint &pos )
{
  // First we want to enable/disable some actions...
  if ( item )
  {
    sapp->sharesUnmount->setEnabled( true );
    sapp->sharesStartFilemanager->setEnabled( true );
  }
  else
  {
    sapp->sharesUnmount->setEnabled( false );
    sapp->sharesStartFilemanager->setEnabled( false );
  }
    
  if ( button == KIconView::LeftButton && !item )
  {
    // We want the Share tab to be cleared, if no item is under the mouse
    // when pressing the left button. So tell the program, that there is 
    // no highlighted item anymore.
    emit info( QString::null );
    // Clear the selection.
    clearSelection();
  }
  else if ( button == KIconView::RightButton )
  {
    m_menu = new KPopupMenu( this );

    m_menu->insertTitle( SmallIcon( "hdd_mount" ), i18n( "Shares" ), Title, 0 );
      
    QIconSet unmountSet( SmallIcon( "hdd_unmount" ), QIconSet::Automatic );
    m_menu->insertItem( unmountSet, i18n( "Unmount" ), this, SLOT( slotUnmountShare() ), i18n( "CTRL+U" ), Unmount, 1 );

    QIconSet unmountallSet( SmallIcon( "gear" ), QIconSet::Automatic );
    m_menu->insertItem( unmountallSet, i18n( "Unmount All" ), this, SLOT( slotUnmountAllShares() ), i18n( "CTRL+N" ), UnmountAll, 2 );

    m_menu->insertSeparator( 3 );

    // Konqueror.
    QIconSet filemanagerSet( SmallIcon( "kfm_home" ), QIconSet::Automatic );
    m_menu->insertItem( filemanagerSet, i18n( "Konqueror" ), this, SLOT( slotOpenFilemanager() ), i18n( "CTRL+K" ), Konqueror, 4 );

    // The program should not send the info( QString::null ) signal here.
    // The behavior of the Share tab would get too weird.
    if ( item && count() != 0 )
    {
      m_menu->setItemEnabled( Unmount, true );
      m_menu->setItemEnabled( Konqueror, true );
      m_menu->setItemEnabled( UnmountAll, true );
    }
    else if ( !item && count() != 0 )
    {
      m_menu->setItemEnabled( Unmount, false );
      m_menu->setItemEnabled( Konqueror, false );
    }
    else
    {
      m_menu->setItemEnabled( Unmount, false );
      m_menu->setItemEnabled( Konqueror, false );
      m_menu->setItemEnabled( UnmountAll, false );
    }

    m_menu->exec( pos, 0 );  
  }
}


void Smb4KShareWidget::slotSelectionChanged( QIconViewItem *item )
{
  // The info about the currently highlighted share has to be 
  // passed to the program. Do this now.
  // When NO share is under the mouse, the signal will be send
  // by slotMouseButtonPressed().
  if ( item )
  {
    if ( m_mountpoint )
      emit info( m_shares.grep( item->text(), true ).first() );
    else
      emit info( m_shares.grep( item->key(), true ).first() );

	  // Enable the actions:
		sapp->sharesUnmount->setEnabled( true );
		sapp->sharesStartFilemanager->setEnabled( true );
  }
}


/****************************************************************************
   External slots.
****************************************************************************/


void Smb4KShareWidget::slotImported( const QStringList &data )
{
  if ( !data.isEmpty() && !m_read )
  {
	  // The old list is replaced by the new one:
		m_shares = data;

		// First remove those, that are not mounted anymore:
		for ( QIconViewItem *item = this->firstItem(); item; item = item->nextItem() )
		{
		  if ( !m_mountpoint && m_shares.grep( item->key(), true ).count() == 0 )
      {
			  delete item;
        this->sort( this->sortDirection() );
      }
			else if ( m_mountpoint && m_shares.grep( item->text(), true ).count() == 0 )
      {
			  delete item;
        this->sort( this->sortDirection() );
      }
		  else
			  continue;
		}

		// Now process the entries in the list:
		for ( QStringList::ConstIterator it = m_shares.begin(); it != m_shares.end(); ++it )
		{
      QString share = (*it).section( ":", 0, 0 ).stripWhiteSpace();
      QString path = (*it).section( ":", 1, 1 ).stripWhiteSpace();
      QString share_uid = (*it).section( ":", 2, 2 ).stripWhiteSpace();
      QString share_gid = (*it).section( ":", 3, 3 ).stripWhiteSpace();

			// Check, whether the share is already in the list. Look for the
			// mount point.
			bool alreadyInView = false;
		  for ( QIconViewItem *item = this->firstItem(); item; item = item->nextItem() )
		  {
		    if ( !m_mountpoint && item->key() == path )
			    alreadyInView = true;
			  else if ( m_mountpoint && item->text() == path )
			    alreadyInView = true;
		    else
			    continue;
		  }

      if ( !alreadyInView )
			{
        // Test whether the imported share belongs to the user.
        bool externalMount = false;
        if ( share_uid.toInt() != m_uid && share_gid.toInt() != m_gid )
        externalMount = true;

        if ( !m_mountpoint )
        {
          if ( m_display == "icons" && !m_showExternal && externalMount )
            continue;
          else if ( m_display == "icons" )
          {
            Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, share, "hdd_mount", m_display, externalMount );
            item->setKey( path );
          }

          if ( m_display == "list" && !m_showExternal && externalMount )
            continue;
          else if ( m_display == "list" )
          {
            Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, share, "hdd_mount", m_display, externalMount );
            item->setKey( path );
          }
        }
        else
        {
          if ( m_display == "icons" && !m_showExternal && externalMount )
            continue;
          else if ( m_display == "icons" )
          {
            Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, path, "hdd_mount", m_display, externalMount );
            item->setKey( share );
          }

          if ( m_display == "list" && !m_showExternal && externalMount )
            continue;
          else if ( m_display == "list" )
          {
            Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, path, "hdd_mount", m_display, externalMount );
            item->setKey( share );
          }
        }
        // Do the sorting:
        this->sort( this->sortDirection() );
		  }
		}
  }
  else if ( data.isEmpty() && !m_read )
  {
    clear();
    emit itemUnmounted();
  }

	// Enable actions:
	if ( count() != 0 && !sapp->sharesUnmountAll->isEnabled() )
	  sapp->sharesUnmountAll->setEnabled( true );
}


void Smb4KShareWidget::slotMounted( const QString &data )
{
  // We know, that only the first entry of the string list is full:
  QString path = data.section( ":", 1, 1 ).stripWhiteSpace();
  QString share = data.section( ":", 0, 0 ).stripWhiteSpace();

  if ( !m_mountpoint )
  {
    Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, share, "hdd_mount", m_display, false );
    item->setKey( path );
  }
  else
  {
    Smb4KShareWidgetItem *item = new Smb4KShareWidgetItem( this, path, "hdd_mount", m_display, false );
    item->setKey( share );
  }
  
  // Append the share to the string list.
  m_shares.append( QString( "%1:%2:%3" ).arg( share ).arg( path ).arg( data.section( ":", 2, 3 ).stripWhiteSpace() ) );
  
  // Do the sorting:
  this->sort( this->sortDirection() );

	// Enable actions:
  sapp->sharesUnmountAll->setEnabled( true );
}


void Smb4KShareWidget::slotUnmounted( const QString &share )
{
  // Remove the share from the list.
  m_shares.remove( share );
       
  delete currentItem();
   
  // Do the sorting:
  this->sort( this->sortDirection() );
    
  emit itemUnmounted();

	// Enabled/disable actions:
	if ( count() != 0 )
	{
	  sapp->sharesUnmount->setEnabled( false );
		sapp->sharesUnmountAll->setEnabled( true );
		sapp->sharesStartFilemanager->setEnabled( false );
	}
	else
	{
	  sapp->sharesUnmount->setEnabled( false );
		sapp->sharesUnmountAll->setEnabled( false );
		sapp->sharesStartFilemanager->setEnabled( false );
	}
}



void Smb4KShareWidget::slotUnmountedAll( const QStringList &shares )
{
  if ( shares.isEmpty() )
  {
    m_shares.clear();
    clear();
  }
  else
  {
    // Only those shares remain in the list view, that are listed in the 
    // passed string list.
    uint index( 0 );
    while ( index < m_shares.count() )
    {
      QString mp = m_shares[index].section( ":", -3, -3 );
      
      if ( shares.grep( mp, true ).isEmpty() )
      {
        if ( m_mountpoint )
        {
          QIconViewItem *item = findItem( mp, ExactMatch|CaseSensitive );
          
          if ( item )
            delete item;
        }
        else
        {
          QIconViewItem *item;
          for ( QIconViewItem *test = this->firstItem(); test; test = test->nextItem() )
          {
            if ( test->key() == mp )
              item = test;
          }
          
          if ( item )
            delete item;
        }
      }
      
      index++;
    }
    // Sort the remaining items.
    this->sort( this->sortDirection() );  
    
    // Replace the old list of mounted shares by the one passed by the mounter:
    m_shares = shares;
  }
  
  emit itemUnmounted();

	// Disable selections:
  sapp->sharesUnmount->setEnabled( false );
	sapp->sharesUnmountAll->setEnabled( false );
	sapp->sharesStartFilemanager->setEnabled( false );
}


void Smb4KShareWidget::slotReceivedError( const QString &output, int errorValue )
{
  if ( errorValue == Smb4KMounter::AlreadyMounted )
    KMessageBox::sorry( this, i18n( "This share has already been mounted!" ), QString::null, KMessageBox::Notify );
  else
    KMessageBox::detailedError( this, i18n( "An error occurred." ), output, QString::null, KMessageBox::Notify );
}


/***************************************************************************
   Slots for the menu.
***************************************************************************/


void Smb4KShareWidget::slotOpenFilemanager()
{
  if ( currentItem() )
  {
    if ( m_mountpoint )
      kapp->invokeBrowser( currentItem()->text() );
    else
      kapp->invokeBrowser( currentItem()->key() );  
  }
}


void Smb4KShareWidget::slotUnmountShare()
{
  if ( currentItem() )
  {
    m_config->setGroup( "Appearance" );
    if ( m_config->readBoolEntry( "Show Mount Point", false ) )
      sapp->mounter()->unmountShare( currentItem()->text() );
    else
      sapp->mounter()->unmountShare( currentItem()->key() );
  }
}


void Smb4KShareWidget::slotUnmountAllShares()
{
  sapp->mounter()->unmountAllShares();
}


/***************************************************************************
 *                                                                         *
 *   The Smb4KShareWidgetItem class.                                       *
 *                                                                         *
 ***************************************************************************/


Smb4KShareWidgetItem::Smb4KShareWidgetItem( Smb4KShareWidget *parent, const QString &text, const QString &iconName, const QString &alignment,  bool external ) : KIconViewItem( parent, text )
{
  m_externalMount = external;
  
  if ( !m_externalMount && alignment == "icons" )
    setPixmap( DesktopIcon( iconName ) );
  else if ( !m_externalMount && alignment == "list" )
    setPixmap( SmallIcon( iconName ) );
  else if ( m_externalMount && alignment == "icons" )
    setPixmap( kapp->iconLoader()->loadIcon( iconName, KIcon::Desktop, 0, KIcon::DisabledState, 0L, false ) );
  else if ( m_externalMount && alignment == "list" )
    setPixmap( kapp->iconLoader()->loadIcon( iconName, KIcon::Small, 0, KIcon::DisabledState, 0L, false ) );
}


Smb4KShareWidgetItem::~Smb4KShareWidgetItem()
{
}


/****************************************************************************
   Paints an item.
****************************************************************************/


void Smb4KShareWidgetItem::paintItem( QPainter *p, const QColorGroup &cg )
{
  QColorGroup colorgrp( cg );
  
  if ( m_externalMount )
  {
/*
    // Change the font:
    QFont f( p->font() );
    f.setItalic( true );
    p->setFont( f );
*/
    colorgrp.setColor( QColorGroup::Text, Qt::gray );
  }
    
  QIconViewItem::paintItem( p, colorgrp );
}


#include "smb4ksharewidget.moc"

